#ifndef F_CPU
	#define F_CPU 12000000
#endif

#include <stdlib.h>
#include <string.h>
#include <avr/io.h>

#include <util/delay.h>


#include "main.h"
#include "file.h"
#include "fat.h"						
#include "hardware.h"

#include "i2clcd.h"

#define ADDR_LM75 	0x92
#define ADDR_RTC 	0xA0

#define FROMBCD(x)      (((x) >> 4) * 10 + ((x) & 0xf))
#define TOBCD(x)        (((x) / 10 * 16) + ((x) % 10))



typedef struct {uint8_t hs, ss, mm, hh, dd, mt, yy, wd, mez, mesz;} date_time_t;

date_time_t rtc;

int8_t h_byte;		// Hight-Byte LM75
uint8_t l_byte;	// Low-Byte LM75

//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}

//************************************************************************
uint8_t lm75_init(void)
{
	if (i2c_start(ADDR_LM75+I2C_WRITE)) return 1;
	i2c_write(0);							// Funktion Temperatur lesen
	i2c_stop();
	lcd_wait_ms(100);						// etwas warten bevor es weiter geht...
	return 0;
}

//************************************************************************
uint8_t lm75_read(void)
{
	if (i2c_start(ADDR_LM75+I2C_READ)) return 1;
	h_byte = i2c_readAck();
	l_byte = i2c_readNak();
	i2c_stop();
	return 0;
}

//***********************************************************	
uint8_t rtc_read(void)
{
	if (i2c_start(ADDR_RTC+I2C_WRITE)) return 1;
	i2c_write(1);					// Adresse 1
	i2c_stop();
	lcd_wait_ms(10);
	if (i2c_start(ADDR_RTC+I2C_READ)) return 1;
	rtc.hs = i2c_readAck();
	rtc.ss = i2c_readAck();
	rtc.mm = i2c_readAck();
	rtc.hh = i2c_readNak();
	rtc.dd = i2c_readNak();
	rtc.mt = i2c_readNak();
	i2c_stop();
	rtc.hs = FROMBCD(rtc.hs);
	rtc.ss = FROMBCD(rtc.ss);
	rtc.mm = FROMBCD(rtc.mm);
	rtc.hh = FROMBCD(rtc.hh);
	rtc.dd = FROMBCD(rtc.dd);
	rtc.mt = FROMBCD(rtc.mt);
	lcd_wait_ms(10);
	return 0;
}

//***********************************************************	
uint8_t rtc_write(unsigned char adr, unsigned char val)
{
	if (i2c_start(ADDR_RTC+I2C_WRITE)) return 1;
	i2c_write(adr);					// Adresse
	i2c_write(val);					// Wert
	i2c_stop();
	return 0;
}

//***********************************************************	
void ItoA(int16_t z, char* Buffer)
{
  int i = 0;
  int j;
  char tmp;
  unsigned u;
  
    // Zahl negativ?
    if( z < 0 ) {
      Buffer[0] = '-';
      Buffer++;
      u = ( (unsigned)-(z+1) ) + 1; 
    }
    else { 
      u = (unsigned)z;
    }
    // die einzelnen Stellen der Zahl berechnen
    do {
      Buffer[i++] = '0' + u % 10;
      u /= 10;
    } while( u > 0 );
 
    // den String in sich spiegeln
    for( j = 0; j < i / 2; ++j ) {
      tmp = Buffer[j];
      Buffer[j] = Buffer[i-j-1];
      Buffer[i-j-1] = tmp;
    }
    Buffer[i] = '\0';
}

//************************************************************************
//************************************************************************
//************************************************************************
int main(void){

  
	i2c_init();
	lcd_init();
	lm75_init();

	lcd_backlight(ON);

	// SD-Card initialisieren
	while (mmc_init() !=0){ 		
		lcd_printlc(1,1,(unsigned char *)"keine SD-Card!");	
	}  
	
	lcd_command(LCD_CLEAR);	

	// FAT auf SD-Card initialisieren
	if(fat_initfat()!=0){				
		lcd_printlc(1,1,(unsigned char *)"keine FAT?!");	
	} else {
		lcd_command(LCD_CLEAR);	
		lcd_printlc(1,1,(unsigned char *)"SD/FAT ok!");
		
		my_delay_ms(1000);	
		lcd_command(LCD_CLEAR);

		// und auf gehts....
		char datei[12]="lm75.log";
		fat_str(datei);
		char s[10];
		char lcd_str[17];
		int i = 0;

		while (1) {
			
			rtc_read();
			lm75_read();
			lcd_str[0]='\0';
			
			// uptime aus RTC
			ItoA(rtc.hh, &s);
			strcat(lcd_str, s);
			strcat(lcd_str, ":");
			ItoA(rtc.mm, &s);
			strcat(lcd_str, s);
			strcat(lcd_str, ":");
			ItoA(rtc.ss, &s);
			strcat(lcd_str, s);
			strcat(lcd_str, ";");
			
			// Temperatur
			if (h_byte >= 128) {
				//negativer Temperaturwert
				strcat(lcd_str, "-");

			} else {
				//positiver Temperaturwert
				strcat(lcd_str, "+");

			}
			ItoA(h_byte, &s);
			strcat(lcd_str, s);
			strcat(lcd_str, ",");
			if (l_byte >= 128) {
				strcat(lcd_str, "5");
			} else {
				strcat(lcd_str, "0");
			}
			
			// Daten auf SD-Card schreiben
			ffopen( datei );
			ffseek(file.length);
			ffwrites((char*)lcd_str);
			ffwrite(0x0D);
			ffwrite(0x0A);
			ffclose();
			
			// ... und Ausgabe auf LCD
			lcd_command(LCD_CLEAR);
			ItoA(i, &s);
			lcd_printlr(2, 1, (unsigned char *)s);
			lcd_printlr(1, 1, (unsigned char *)lcd_str);
			i++;
			
			my_delay_ms(5000);
		}
	}
}
